/* This is a modified version of dsputil_alpha.c from ffmpeg.
 * Modifications by Mns Rullgrd <mru@users.sf.net>
 */

/*
 * Alpha optimized DSP utils
 * Copyright (c) 2002 Falk Hueffner <falk@debian.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <inttypes.h>

#include "mpeg2.h"
#include "mpeg2_internal.h"
#include "asm_alpha.h"

static inline uint64_t avg2_no_rnd(uint64_t a, uint64_t b)
{
    return (a & b) + (((a ^ b) & BYTE_VEC(0xfe)) >> 1);
}

static inline uint64_t avg2(uint64_t a, uint64_t b)
{
    return (a | b) - (((a ^ b) & BYTE_VEC(0xfe)) >> 1);    
}

static inline uint64_t avg4(uint64_t l1, uint64_t l2, uint64_t l3, uint64_t l4)
{
    uint64_t r1 = ((l1 & ~BYTE_VEC(0x03)) >> 2)
	      + ((l2 & ~BYTE_VEC(0x03)) >> 2)
	      + ((l3 & ~BYTE_VEC(0x03)) >> 2)
	      + ((l4 & ~BYTE_VEC(0x03)) >> 2);
    uint64_t r2 = ((  (l1 & BYTE_VEC(0x03))
		  + (l2 & BYTE_VEC(0x03))
		  + (l3 & BYTE_VEC(0x03))
		  + (l4 & BYTE_VEC(0x03))
		  + BYTE_VEC(0x02)) >> 2) & BYTE_VEC(0x03);
    return r1 + r2;
}

#define OP8(LOAD, STORE)			\
    do {					\
        STORE(LOAD(pixels), block);		\
        pixels += line_size;			\
        block += line_size;			\
    } while (--h)

#define OP16(LOAD, STORE)			\
    do {					\
        STORE(LOAD(pixels), block);		\
        STORE(LOAD(pixels+8), block+8);		\
        pixels += line_size;			\
        block += line_size;			\
    } while (--h)

#define OP_X8(LOAD, STORE)                                      \
    do {                                                        \
        uint64_t pix1, pix2;                                    \
                                                                \
        pix1 = LOAD(pixels);                                    \
        pix2 = pix1 >> 8 | ((uint64_t) pixels[8] << 56);        \
        STORE(AVG2(pix1, pix2), block);                         \
        pixels += line_size;                                    \
        block += line_size;                                     \
    } while (--h)

#define OP_X16(LOAD, STORE)					\
    do {							\
        uint64_t pix1, pix2, p3, p4;				\
								\
        pix1 = LOAD(pixels);					\
        p3 = LOAD(pixels+8);					\
        pix2 = pix1 >> 8 | ((uint64_t) pixels[8] << 56);	\
        p4 = p3 >> 8 | ((uint64_t) pixels[16] << 56);		\
        STORE(AVG2(pix1, pix2), block);				\
        STORE(AVG2(p3, p4), block+8);				\
        pixels += line_size;					\
        block += line_size;					\
    } while (--h)

#define OP_Y8(LOAD, STORE)                      \
    do {                                        \
        uint64_t pix = LOAD(pixels);            \
        do {                                    \
            uint64_t next_pix;                  \
                                                \
            pixels += line_size;                \
            next_pix = LOAD(pixels);            \
            STORE(AVG2(pix, next_pix), block);  \
            block += line_size;                 \
            pix = next_pix;                     \
        } while (--h);                          \
    } while (0)

#define OP_Y16(LOAD, STORE)			\
    do {					\
        uint64_t pix = LOAD(pixels);		\
        uint64_t pix1 = LOAD(pixels+8);		\
        do {					\
            uint64_t next_pix, np1;		\
						\
            pixels += line_size;		\
            next_pix = LOAD(pixels);		\
            np1 = LOAD(pixels+8);		\
            STORE(AVG2(pix, next_pix), block);	\
            STORE(AVG2(pix1, np1), block+8);	\
            block += line_size;			\
            pix = next_pix;			\
            pix1 = np1;				\
        } while (--h);				\
    } while (0)

#define OP_XY8(LOAD, STORE)                                                 \
    do {                                                                    \
        uint64_t pix1 = LOAD(pixels);                                       \
        uint64_t pix2 = pix1 >> 8 | ((uint64_t) pixels[8] << 56);           \
        uint64_t pix_l = (pix1 & BYTE_VEC(0x03))                            \
                       + (pix2 & BYTE_VEC(0x03));                           \
        uint64_t pix_h = ((pix1 & ~BYTE_VEC(0x03)) >> 2)                    \
                       + ((pix2 & ~BYTE_VEC(0x03)) >> 2);                   \
                                                                            \
        do {                                                                \
            uint64_t npix1, npix2;                                          \
            uint64_t npix_l, npix_h;                                        \
            uint64_t avg;                                                   \
                                                                            \
            pixels += line_size;                                            \
            npix1 = LOAD(pixels);                                           \
            npix2 = npix1 >> 8 | ((uint64_t) pixels[8] << 56);              \
            npix_l = (npix1 & BYTE_VEC(0x03))                               \
                   + (npix2 & BYTE_VEC(0x03));                              \
            npix_h = ((npix1 & ~BYTE_VEC(0x03)) >> 2)                       \
                   + ((npix2 & ~BYTE_VEC(0x03)) >> 2);                      \
            avg = (((pix_l + npix_l + AVG4_ROUNDER) >> 2) & BYTE_VEC(0x03)) \
                + pix_h + npix_h;                                           \
            STORE(avg, block);                                              \
                                                                            \
            block += line_size;                                             \
            pix_l = npix_l;                                                 \
            pix_h = npix_h;                                                 \
        } while (--h);                                                      \
    } while (0)

#define OP_XY16(LOAD, STORE)						    \
    do {								    \
        uint64_t pix1 = LOAD(pixels);					    \
        uint64_t pix2 = pix1 >> 8 | ((uint64_t) pixels[8] << 56);	    \
        uint64_t pix_l = (pix1 & BYTE_VEC(0x03))			    \
                       + (pix2 & BYTE_VEC(0x03));			    \
        uint64_t pix_h = ((pix1 & ~BYTE_VEC(0x03)) >> 2)		    \
                       + ((pix2 & ~BYTE_VEC(0x03)) >> 2);		    \
									    \
        uint64_t pix3 = LOAD(pixels+8);					    \
        uint64_t pix4 = pix3 >> 8 | ((uint64_t) pixels[16] << 56);	    \
        uint64_t pix_l3 = (pix3 & BYTE_VEC(0x03))			    \
                       + (pix4 & BYTE_VEC(0x03));			    \
        uint64_t pix_h3 = ((pix3 & ~BYTE_VEC(0x03)) >> 2)		    \
                       + ((pix4 & ~BYTE_VEC(0x03)) >> 2);		    \
									    \
        do {								    \
            uint64_t npix1, npix2, npix3, npix4;			    \
            uint64_t npix_l, npix_h, npl3, nph3;			    \
            uint64_t avg, avg3;						    \
									    \
            pixels += line_size;					    \
            npix1 = LOAD(pixels);					    \
            npix2 = npix1 >> 8 | ((uint64_t) pixels[8] << 56);		    \
            npix_l = (npix1 & BYTE_VEC(0x03))				    \
                   + (npix2 & BYTE_VEC(0x03));				    \
            npix_h = ((npix1 & ~BYTE_VEC(0x03)) >> 2)			    \
                   + ((npix2 & ~BYTE_VEC(0x03)) >> 2);			    \
            avg = (((pix_l + npix_l + AVG4_ROUNDER) >> 2) & BYTE_VEC(0x03)) \
                + pix_h + npix_h;					    \
            STORE(avg, block);						    \
									    \
            npix3 = LOAD(pixels+8);					    \
            npix4 = npix3 >> 8 | ((uint64_t) pixels[16] << 56);		    \
            npl3 = (npix3 & BYTE_VEC(0x03))				    \
                   + (npix4 & BYTE_VEC(0x03));				    \
            nph3 = ((npix3 & ~BYTE_VEC(0x03)) >> 2)			    \
                   + ((npix4 & ~BYTE_VEC(0x03)) >> 2);			    \
            avg3 = (((pix_l3 + npl3 + AVG4_ROUNDER) >> 2) & BYTE_VEC(0x03)) \
                + pix_h3 + nph3;					    \
            STORE(avg3, block+8);					    \
									    \
            block += line_size;						    \
            pix_l = npix_l;						    \
            pix_h = npix_h;						    \
            pix_l3 = npl3;						    \
            pix_h3 = nph3;						    \
        } while (--h);							    \
    } while (0)

#define MAKE_OP(OPNAME, SUFF, OPKIND, STORE)				  \
static void MC_ ## OPNAME ## SUFF ## _alpha				  \
        (uint8_t *__restrict__ block, const uint8_t *__restrict__ pixels, \
         int line_size, int h)						  \
{									  \
    if ((size_t) pixels & 0x7) {					  \
        OPKIND(uldq, STORE);						  \
    } else {								  \
        OPKIND(ldq, STORE);						  \
    }									  \
}

#define PIXOP(OPNAME, STORE)			\
    MAKE_OP(OPNAME, _8,   OP8,    STORE)	\
    MAKE_OP(OPNAME, _x8,  OP_X8,  STORE)	\
    MAKE_OP(OPNAME, _y8,  OP_Y8,  STORE)	\
    MAKE_OP(OPNAME, _xy8, OP_XY8, STORE)	\
    MAKE_OP(OPNAME, _16,  OP16,   STORE)	\
    MAKE_OP(OPNAME, _x16, OP_X16, STORE)	\
    MAKE_OP(OPNAME, _y16, OP_Y16, STORE)	\
    MAKE_OP(OPNAME, _xy16,OP_XY16,STORE)

/* Rounding primitives.  */
#define AVG2 avg2
#define AVG4 avg4
#define AVG4_ROUNDER BYTE_VEC(0x02)
#define STORE(l, b) stq(l, b)
PIXOP(put, STORE);

#undef STORE
#define STORE(l, b) stq(AVG2(l, ldq(b)), b);
PIXOP(avg, STORE);


MOTION_COMP_EXTERN(alpha);
