#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <dlfcn.h> /* GLIBC specific. Exists under cygwin too! */

#include "config.h"
#include "help_mp.h"

#include "bswap.h"

#include "vd_internal.h"
#include "codecs_ld.h"
#include "../postproc/postprocess.h"

#define MBC 128
#define MBR 96

static vd_info_t info = {
	"FFmpeg's libavcodec codec family",
	"ffmpeg",
	"A'rpi",
	"http://ffmpeg.sf.net",
	"native codecs"
};

LIBVD_EXTERN(ffmpeg)

#include "interface/ffmpeg/avcodec.h"
#include "libvo/video_out.h"

#undef ENABLE_DR1 /* too buggy and slow */
/* NOTE: DR1 will be possible only with VIDIX + BUSMASTERING after implementing
of HW OSD support when they will be painted into independed buffers. */

static int vcodec_inited=0;
#ifdef ENABLE_DR1
static int do_dr1=0;
#endif
/* to set/get/query special features/parameters */
static int control(sh_video_t *sh,int cmd,void* arg,...){
    switch(cmd){
    case VDCTRL_QUERY_MAX_PP_LEVEL:
	return GET_PP_QUALITY_MAX;
    case VDCTRL_SET_PP_LEVEL: {
	int quality=*((int*)arg);
	if(quality<0 || quality>GET_PP_QUALITY_MAX) quality=GET_PP_QUALITY_MAX;
	divx_quality = quality;
	return CONTROL_OK;
    }
    }
    return CONTROL_UNKNOWN;
}

unsigned (*avcodec_version_ptr)(void);
unsigned (*avcodec_build_ptr)(void);
void (*avcodec_init_ptr)(void);
void (*avcodec_register_all_ptr)(void);
AVCodec * (*avcodec_find_decoder_by_name_ptr)(const char *name);
int (*avcodec_open_ptr)(AVCodecContext *avctx, AVCodec *codec);
int (*avcodec_close_ptr)(AVCodecContext *avctx);
int (*avcodec_decode_video_ptr)(AVCodecContext *avctx, AVPicture *picture, 
                         int *got_picture_ptr,
                         UINT8 *buf, int buf_size);
AVCodecContext* (*avcodec_alloc_context_ptr)(void);


static void *dll_handle;
static int load_dll(const char *libname)
{
  if(!(dll_handle=ld_codec(libname,"http://ffmpeg.sf.net"))) return 0;
  avcodec_version_ptr = dlsym(dll_handle,"avcodec_version");
  avcodec_build_ptr = dlsym(dll_handle,"avcodec_build");
  avcodec_init_ptr = dlsym(dll_handle,"avcodec_init");
  avcodec_register_all_ptr = dlsym(dll_handle,"avcodec_register_all");
  avcodec_find_decoder_by_name_ptr = dlsym(dll_handle,"avcodec_find_decoder_by_name");
  avcodec_open_ptr = dlsym(dll_handle,"avcodec_open");
  avcodec_close_ptr = dlsym(dll_handle,"avcodec_close");
  avcodec_decode_video_ptr = dlsym(dll_handle,"avcodec_decode_video");
  avcodec_alloc_context_ptr = dlsym(dll_handle,"avcodec_alloc_context");
  return avcodec_version_ptr && avcodec_init_ptr && avcodec_register_all_ptr &&
	 avcodec_find_decoder_by_name_ptr && avcodec_open_ptr && avcodec_close_ptr &&
	 avcodec_decode_video_ptr && avcodec_build_ptr && avcodec_alloc_context_ptr;
}

#define MIN_LIBAVCODEC_VERSION_INT	0x000406
#define MIN_LIBAVCODEC_BUILD		4638

#ifdef ENABLE_DR1
static void get_buffer(struct AVCodecContext *avctx, int width, int height, int pict_type){
    sh_video_t * sh = avctx->opaque;
    mp_image_t* mpi=NULL;
    int flags= MP_IMGFLAG_ACCEPT_STRIDE|MP_IMGFLAG_PRESERVE|MP_IMGFLAG_READABLE;
    int type= MP_IMGTYPE_IPB;
    int align=15;

    if(avctx->pix_fmt == PIX_FMT_YUV410P)
        align=63; //yes seriously, its really needed (16x16 chroma blocks in SVQ1 -> 64x64)

    if(avctx->has_b_frames){
        type= MP_IMGTYPE_IPB;
    }else{
        type= MP_IMGTYPE_IP;
    }
    MSG_DBG2(type== MP_IMGTYPE_IPB ? "using IPB\n" : "using IP\n");

    mpi= mpcodecs_get_image(sh,type, flags,
			(width+align)&(~align), (height+align)&(~align));

    // ok, lets see what did we get:
    avctx->draw_horiz_band= NULL;
    avctx->dr_buffer[0]= mpi->planes[0];
    avctx->dr_buffer[1]= mpi->planes[1];
    avctx->dr_buffer[2]= mpi->planes[2];
    
    if(avctx->dr_stride && avctx->dr_stride !=mpi->stride[0]){
        MSG_ERR("Error: stride changed\n");
    }

    avctx->dr_stride   = mpi->stride[0];
    avctx->dr_uvstride = mpi->stride[1];

    avctx->dr_opaque_frame = mpi;
    avctx->dr_ip_buffer_count=2; //FIXME
}
#endif

static int quant_store[MBR+1][MBC+1];
static int init(sh_video_t *sh){
    unsigned avc_version;
    unsigned avc_build;
    AVCodec *lavc_codec;
    AVCodecContext *ctx;

    if(!vcodec_inited){
      if(!load_dll("libavcodec-0.4.6.so"))
	if(!load_dll("libavcodec.so"))
	{
	    MSG_ERR("Detected error during loading libffmpeg.so! Try to upgrade this codec\n");
	    return 0;
	}
	avc_version = (*avcodec_version_ptr)();
	avc_build = (*avcodec_build_ptr)();
	if(avc_version < LIBAVCODEC_VERSION_INT && avc_build < LIBAVCODEC_BUILD)
	{
	    MSG_ERR("You have wrong version of libavcodec %06X (build %06X)\n."
		   "Right version should be %06X (build %06X)\n",
		    avc_version,avc_build,MIN_LIBAVCODEC_VERSION_INT,MIN_LIBAVCODEC_BUILD);
	    return 0;
	}
        (*avcodec_init_ptr)();
        (*avcodec_register_all_ptr)();
        vcodec_inited=1;
    }
    lavc_codec = (AVCodec *)(*avcodec_find_decoder_by_name_ptr)(sh->codec->dll_name);
    if(!lavc_codec){
	MSG_ERR(MSGTR_MissingLAVCcodec,sh->codec->dll_name);
	return 0;
    }
    
    ctx = sh->context = (*avcodec_alloc_context_ptr)();
    
    ctx->width = sh->disp_w;
    ctx->height= sh->disp_h;
    ctx->quant_store = &quant_store[0][0];
    ctx->qstride = MBC+1;
    MSG_DBG2("libavcodec.size: %d x %d\n",ctx->width,ctx->height);
    if (sh->format == mmioFOURCC('R', 'V', '1', '3'))
	ctx->sub_id = 3;
    /* AVRn stores huffman table in AVI header */
    /* Pegasus MJPEG stores it also in AVI header, but it uses the common
       MJPG fourcc :( */
    if (sh->bih && (sh->bih->biSize != sizeof(BITMAPINFOHEADER)) &&
	(sh->format == mmioFOURCC('A','V','R','n') ||
	sh->format == mmioFOURCC('M','J','P','G')))
    {
	ctx->flags |= CODEC_FLAG_EXTERN_HUFF;
	ctx->extradata_size = sh->bih->biSize-sizeof(BITMAPINFOHEADER);
	ctx->extradata = malloc(ctx->extradata_size);
	memcpy(ctx->extradata, sh->bih+sizeof(BITMAPINFOHEADER),
	    ctx->extradata_size);
    }
#ifdef ENABLE_DR1
    if(lavc_codec->capabilities&CODEC_CAP_DR1)
    {
	do_dr1=1;
        ctx->flags|= CODEC_FLAG_EMU_EDGE | CODEC_FLAG_DR1; 
        ctx->get_buffer_callback= get_buffer;
    }
#endif
    /* open it */
    if ((*avcodec_open_ptr)(ctx, lavc_codec) < 0) {
        MSG_ERR( MSGTR_CantOpenCodec);
        return 0;
    }
    MSG_V("INFO: libavcodec (libffmpeg.so) video codec init OK!\n");
    return mpcodecs_config_vo(sh,sh->disp_w,sh->disp_h,IMGFMT_YV12);
}

// uninit driver
static void uninit(sh_video_t *sh){
    AVCodecContext *ctx = sh->context;
    if ((*avcodec_close_ptr)(sh->context) < 0)
    	    MSG_ERR( MSGTR_CantCloseCodec);
    if (ctx->extradata_size)
	free(ctx->extradata);
    dlclose(dll_handle);
    vcodec_inited=0;
}

//#define FF_SLICES 1
/* 
   Cel1 no slices xp   = VC: 23.2658% VO:  9.2109% (32.47%) (540x262@25)
   Duron		 VC: 18.3419% VO: 13.2822% (31.62%) (720x480@30)
			 VC:  6.3447% VO:  5.5977% (11.94%) (528x288@25)

   Cel1 slices    xp   = VC: 28.5326% VO:  4.0495% (32.58%)
   Duron		 VC: 28.7263% VO:  2.7816% (31.51%)
			 VC:  9.6781% VO:  2.1125% (11.79%)
   
   Cel1 no slices noxp = VC: 23.6481% VO:  5.1853% (28.82%)
   Duron		 VC: 17.6963% VO: 10.4160% (28.12%)
			 VC:  5.4896% VO:  3.2548% (8.74%)
   
   Cel1 slices    noxp = VC: 29.0462% VO:  0.0428% (29.08%)
   Duron		 VC: 27.9116% VO:  0.0323% (27.94%)
			 VC:  8.6082% VO:  0.0212% (8.63%)
   Who need these slices?
 */
#ifdef FF_SLICES
static void draw_slice(struct AVCodecContext *s,
                	UINT8 **src, int linesize,
                	int y, int width, int height){
    int stride[3];

    stride[0]=linesize;
    stride[1]=stride[2]=stride[0]/2;

    vo_draw_slice (src, stride, width, height, 0, y);
}
#endif
// decode a frame
static mp_image_t* decode(sh_video_t *sh,void* data,int len,int flags){
    int got_picture=0;
    int ret;
    AVPicture lavc_picture;
    AVCodecContext *ctx = sh->context;
    mp_image_t* mpi;
    mp_image_t* mpi_pp;

    ctx->opaque=sh;
    if(len<=0) return NULL; // skipped frame

#ifdef FF_SLICES    
    if(!(flags&3)
#ifdef ENABLE_DR1
    &&!do_dr1
#endif
    ) 
    {
	mpi=mpcodecs_get_image(sh, MP_IMGTYPE_EXPORT, MP_IMGFLAG_PRESERVE |
				MP_IMGFLAG_DRAW_CALLBACK,sh->disp_w, sh->disp_h);
	if(mpi && mpi->flags&MP_IMGFLAG_DRAW_CALLBACK) ctx->draw_horiz_band=draw_slice;
    }
#else
    ctx->draw_horiz_band=NULL;
#endif
    ctx->quant_store = &quant_store[0][0];
    ctx->qstride = MBC+1;
    ctx->hurry_up=(flags&3)?((flags&2)?2:1):0;

    ret = (*avcodec_decode_video_ptr)(sh->context, &lavc_picture,
	     &got_picture, data, len);
    
    if(ret<0) MSG_WARN("Error while decoding frame!\n");
    if(!got_picture) return NULL;	// skipped image
#ifndef FF_SLICES   
#ifdef ENABLE_DR1
    if(do_dr1 && ctx->dr_opaque_frame){
        mpi= (mp_image_t*)ctx->dr_opaque_frame;
    }
    else
    {
#endif
	mpi=mpcodecs_get_image(sh, MP_IMGTYPE_EXPORT, MP_IMGFLAG_PRESERVE,
				ctx->width, ctx->height);
	if(!mpi){	// temporary!
	    MSG_ERR("couldn't allocate image for ffmpeg codec\n");
	    return NULL;
	}
    
	mpi->planes[0]=lavc_picture.data[0];
	mpi->planes[1]=lavc_picture.data[1];
	mpi->planes[2]=lavc_picture.data[2];
	mpi->stride[0]=lavc_picture.linesize[0];
	mpi->stride[1]=lavc_picture.linesize[1];
	mpi->stride[2]=lavc_picture.linesize[2];
#ifdef ENABLE_DR1
    }
#endif
    if(sh->codec->outfmt[sh->outfmtidx] == IMGFMT_I420 ||
       sh->codec->outfmt[sh->outfmtidx] == IMGFMT_IYUV)
    {
	uint8_t *tmp;
	unsigned ls;
	tmp=mpi->planes[2];
	mpi->planes[2]=mpi->planes[1];
	mpi->planes[1]=tmp;
	ls=mpi->stride[2];
	mpi->stride[2]=mpi->stride[1];
	mpi->stride[1]=ls;
    }
    if(ctx->pix_fmt==PIX_FMT_YUV422P){
	mpi->stride[1]*=2;
	mpi->stride[2]*=2;
    }
    if(divx_quality && !(flags&3))
    {
	mpi_pp=mpcodecs_get_image(sh, MP_IMGTYPE_TEMP,
				MP_IMGFLAG_PRESERVE | MP_IMGFLAG_READABLE | MP_IMGFLAG_ACCEPT_STRIDE,
				(sh->disp_w+15)&~15, (sh->disp_h+15)&~15);
	if(mpi_pp)
	{
	    postprocess(mpi->planes,mpi->stride[0],
			mpi_pp->planes,mpi_pp->stride[0],
			sh->disp_w,sh->disp_h,
			&quant_store[0][0],MBC+1,getPpModeForQuality(divx_quality));
	}
	mpi = mpi_pp;
    }
#endif
    return mpi;
}

