/*
 *  video_out.h
 *
 *      Copyright (C) Aaron Holtzman - Aug 1999
 *	Strongly modified, most parts rewritten: A'rpi/ESP-team - 2000-2001
 *
 */
 
#ifndef __VIDEO_OUT_H
#define __VIDEO_OUT_H 1

#include <inttypes.h>
#include <stdarg.h>

#include "font_load.h"
#include "img_format.h"
#include "../vidix/vidix.h"
#include "../mp_image.h"

#define VO_EVENT_EXPOSE 1
#define VO_EVENT_RESIZE 2
#define VO_EVENT_KEYPRESS 4
#define VO_EVENT_FORCE_UPDATE 0x80000000

/* takes a pointer to a vo_vaa_s struct */
#define VOCTRL_QUERY_VAA 1
/* takes a pointer to uint32_t fourcc */
#define VOCTRL_QUERY_FORMAT 2
/* signal a device reset seek */
#define VOCTRL_RESET 3
/* used to switch to fullscreen. Must affect window only (not surfaces) */
#define VOCTRL_FULLSCREEN 4
/* user wants to have screen shot. (currently without args)*/
#define VOCTRL_SCREENSHOT 5
/* signal a device pause */
#define VOCTRL_PAUSE 6
/* start/resume playback */
#define VOCTRL_RESUME 7
/* libmpcodecs direct rendering: */
#define VOCTRL_GET_IMAGE 8
/* Input's events : arg = (uint32_t *)&event_type. Must return at least VO_EVENT_RESIZE */
#define VOCTRL_CHECK_EVENTS 9
/* multibuffering */
#define VOCTRL_GET_NUM_FRAMES 10
/* There are no drivers which implement this call */
#define VOCTRL_DRAW_OSD 12
/* Flush frame from RAM into video memory */
#define VOCTRL_FLUSH_FRAME 13
/* Draw-slice callback for non-DRI drivers */
#define VOCTRL_DRAW_SLICE 14

#define VO_TRUE		1
#define VO_FALSE	0
#define VO_ERROR	-1
#define VO_NOTAVAIL	-2
#define VO_NOTIMPL	-3

#define VOFLAG_FULLSCREEN	0x01
#define VOFLAG_MODESWITCHING	0x02
#define VOFLAG_SWSCALE		0x04
#define VOFLAG_FLIPPING		0x08

typedef struct vo_info_s
{
        /* driver name ("Matrox Millennium G200/G400" */
        const char *name;
        /* short name (for config strings) ("mga") */
        const char *short_name;
        /* author ("Aaron Holtzman <aholtzma@ess.engr.uvic.ca>") */
        const char *author;
        /* any additional comments */
        const char *comment;
} vo_info_t;

/* Direct access to BES */
typedef struct bes_da_s
{
	vidix_rect_t	dest;           /* This field should be filled by x,y,w,h
					   from vidix:src but pitches from
					   vidix:dest */
	int		flags;          /* Probably will work only when flag == 0 */
	/* memory model */
	unsigned	frame_size;		/* destination frame size */
	unsigned	num_frames;		/* number of available frames */
	unsigned	offsets[VID_PLAY_MAXFRAMES]; /* relative offset of each frame from begin of video memory */
	vidix_yuv_t	offset;			/* relative offsets within frame for yuv planes */
	void*		dga_addr;		/* linear address of BES */
}bes_da_t;

/*
   Video Accelearted Architecture.
   Every field of this structure can be set to NULL that means that
   features is not supported
*/
typedef struct vo_vaa_s
{
	uint32_t    flags; /* currently undefined */
		/*
		 * Query Direct Access to BES
		 * info - information to be filled
		 * returns: 0 on success errno on error.
		 */
	int  (*query_bes_da)(bes_da_t *info);
	int  (*get_video_eq)(vidix_video_eq_t *info);
	int  (*set_video_eq)(const vidix_video_eq_t *info);
	int  (*get_num_fx)(unsigned *info);
	int  (*get_oem_fx)(vidix_oem_fx_t *info);
	int  (*set_oem_fx)(const vidix_oem_fx_t *info);
	int  (*set_deint)(const vidix_deinterlace_t *info);
}vo_vaa_t;

/* Misc info to tuneup vo driver */
typedef struct vo_tune_info_s
{
	int	pitch[3]; /* Should be 0 if unknown else power of 2 */
}vo_tune_info_t;

/* slice description for drivers which don't support DRI */
typedef struct vo_slice_s
{
	uint8_t		*src[4];
	unsigned	stride[4];
	unsigned	x,y,w,h;
}vo_slice_t;

typedef struct vo_query_fourcc_s
{
	uint32_t	fourcc;
	unsigned	w,h;
}vo_query_fourcc_t;

typedef struct vo_functions_s
{
	/*
	 * Preinitializes driver (real INITIALIZATION)
	 *   arg - currently it's vo_subdevice
	 *   returns: zero on successful initialization, non-zero on error.
	 */
	uint32_t (*preinit)(const char *arg);
        /*
         * Initialize (means CONFIGURE) the display driver.
	 * params:
         *   width,height: image source size
	 *   d_width,d_height: size of the requested window size, just a hint
	 *   fullscreen: flag, 0=windowd 1=fullscreen, just a hint
	 *   title: window title, if available
	 *   format: fourcc of pixel format
         * returns : zero on successful initialization, non-zero on error.
         */
        uint32_t (*config)(uint32_t width, uint32_t height, uint32_t d_width,
			 uint32_t d_height, uint32_t fullscreen, char *title,
			 uint32_t format,const vo_tune_info_t *);

	/*
	 * Control interface
	 */
	uint32_t (*control)(uint32_t request, void *data, ...);

        /*
         * Return driver information.
         *   returns : read-only pointer to a vo_info_t structure.
         */
        const vo_info_t* (*get_info)(void);

        /*
         * Blit/Flip buffer to the screen. Must be called after each frame!
         */
        void (*flip_page)(unsigned idx);

        /*
         * Closes driver. Should restore the original state of the system.
         */
        void (*uninit)(void);

} vo_functions_t;

char *vo_format_name(int format);
/******************************************************
* High level VO functions to provide some abstraction *
* level for video out library                         *
\*****************************************************/
extern void		vo_print_help( void );
extern vo_functions_t *	vo_register(const char *driver_name);
extern const vo_info_t*	vo_get_info( void );
extern int		vo_init( const char *subdevice_name);
extern uint32_t 	vo_config(uint32_t width, uint32_t height, uint32_t d_width,
				  uint32_t d_height, uint32_t fullscreen, char *title,
				  uint32_t format,const vo_tune_info_t *);
extern uint32_t		vo_query_vaa( vo_vaa_t* vaa );
extern uint32_t		vo_query_format( uint32_t* fourcc,unsigned src_w,unsigned src_h);
extern uint32_t		vo_reset( void );
extern uint32_t		vo_fullscreen( void );
extern uint32_t		vo_screenshot( void );
extern uint32_t		vo_pause( void );
extern uint32_t		vo_resume( void );
extern uint32_t		vo_get_surface( mp_image_t* mpi );
extern int		vo_check_events( void );
extern uint32_t		vo_get_num_frames( unsigned * );
extern uint32_t		vo_get_frame_num( volatile unsigned * );
extern uint32_t		vo_set_frame_num( volatile unsigned * );
extern uint32_t		vo_get_active_frame( volatile unsigned * );
extern uint32_t		vo_set_active_frame( volatile unsigned * );
extern uint32_t 	vo_draw_frame(uint8_t *src[]);
extern uint32_t 	vo_draw_slice(uint8_t *src[], unsigned stride[], 
				      unsigned w,unsigned h,unsigned x,unsigned y);
extern void		vo_flip_page(void);
extern void		vo_flush_frame(void);
extern void		vo_draw_osd(void);
extern void		vo_uninit( void );

int vo_x11_init(void);

// NULL terminated array of all drivers
extern vo_functions_t* video_out_drivers[];

extern int vo_flags;

// correct resolution/bpp on screen:  (should be autodetected by vo_x11_init())
extern unsigned vo_depthonscreen;
extern unsigned vo_screenwidth;
extern unsigned vo_screenheight;

// requested resolution/bpp:  (-x -y -bpp options)
extern unsigned vo_dx;
extern unsigned vo_dy;
extern unsigned vo_dwidth;
extern unsigned vo_dheight;
extern unsigned vo_dbpp;

extern unsigned vo_old_x;
extern unsigned vo_old_y; 
extern unsigned vo_old_width;
extern unsigned vo_old_height;

extern int vo_doublebuffering;
extern int vo_vsync;
extern int vo_fs;
extern int vo_fsmode;

extern int vo_pts;
extern float vo_fps;

extern char *vo_subdevice;

#endif
