#define DISP

/* 
 * vo_png.c, Portable Network Graphics Renderer for Mplayer
 *
 * Copyright 2001 by Felix Buenemann <atmosfear@users.sourceforge.net>
 *
 * Uses libpng (which uses zlib), so see according licenses.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "config.h"

#ifdef HAVE_LIBPNG_PNG
#include <libpng/png.h>
#else
#include <png.h>
#endif

#include "video_out.h"
#include "video_out_internal.h"

#include "../postproc/swscale.h"
#include "../postproc/rgb2rgb.h"
#include "sub.h"
LIBVO_EXTERN (png)

static vo_info_t vo_info = 
{
	"PNG file",
	"png",
	"Felix Buenemann <atmosfear@users.sourceforge.net>",
	""
};

#define RGB 0
#define BGR 1

extern int verbose;
int z_compression = Z_NO_COMPRESSION;
static unsigned src_width;
static unsigned src_height;
static unsigned src_bpp;
static unsigned image_width;
static unsigned image_height;
static unsigned image_format;
static uint8_t *image_data=NULL;

static int bpp = 24;
static int cspace = RGB;
static int framenum = 0;

static SwsContext * sws = NULL;

struct pngdata {
	FILE * fp;
	png_structp png_ptr;
	png_infop info_ptr;
	enum {OK,ERROR} status;  
};
	
static uint32_t
config(uint32_t width, uint32_t height, uint32_t d_width, uint32_t d_height, uint32_t fullscreen, char *title, uint32_t format,const vo_tune_info_t *info)
{
    UNUSED(fullscreen);
    UNUSED(title);
    UNUSED(info);
    src_width = width;
    src_height = height;
    image_width = (d_width+7)&~7;
    image_height = d_height;
    image_format = format;
    sws = getSwsContextFromCmdLine(width,height,format,image_width,image_height,IMGFMT_BGR24);
    if(!sws)
    {
	MSG_ERR("vo_png: Can't initialize SwScaler\n");
	return -1;
    }
    if(!(image_data = malloc(image_width*image_height*3)))
    {
	MSG_ERR("vo_png: Can't allocate temporary buffer\n");
	return -1;
    }
    switch(format) {
	case IMGFMT_RGB32:
	case IMGFMT_BGR32: src_bpp = 32; break;
	case IMGFMT_RGB24:
	case IMGFMT_BGR24: src_bpp = 24; break;
	case IMGFMT_RGB15:
	case IMGFMT_BGR15: src_bpp = 15; break;
	default:
	case IMGFMT_RGB16:
	case IMGFMT_BGR16: src_bpp = 16; break;
    }
    switch(format) {
	case IMGFMT_RGB24:
	     bpp = 24;
	     cspace = RGB;
	break;     
	default:
	     bpp = 24;
	     cspace = BGR;
	break;
    }		
    
    if((z_compression >= 0) && (z_compression <= 9)) {
	    if(z_compression == 0) {
		    MSG_HINT("PNG Warning: compression level set to 0, compression disabled!\n");
		    MSG_HINT("PNG Info: Use the -z <n> switch to set compression level from 0 to 9.\n");
		    MSG_HINT("PNG Info: (0 = no compression, 1 = fastest, lowest - 9 best, slowest compression)\n");
	    }	    
    }
    else {	    	    
	    MSG_WARN("PNG Warning: compression level out of range setting to 1!\n");
	    MSG_WARN("PNG Info: Use the -z <n> switch to set compression level from 0 to 9.\n");
	    MSG_WARN("PNG Info: (0 = no compression, 1 = fastest, lowest - 9 best, slowest compression)\n");
	    z_compression = Z_BEST_SPEED;
    }
    
    MSG_V("PNG Compression level %i\n", z_compression);   
	  	
    return 0;
}

static const vo_info_t*
get_info(void)
{
    return &vo_info;
}


struct pngdata create_png (char * fname)
{
    struct pngdata png;
    
    png.png_ptr = png_create_write_struct
       (PNG_LIBPNG_VER_STRING, NULL,
        NULL, NULL);
    png.info_ptr = png_create_info_struct(png.png_ptr);
   
    if (!png.png_ptr) {
       MSG_V("PNG Failed to init png pointer\n");
       png.status = ERROR;
       return png;
    }   
    
    if (!png.info_ptr) {
       MSG_V("PNG Failed to init png infopointer\n");
       png_destroy_write_struct(&png.png_ptr,
         (png_infopp)NULL);
       png.status = ERROR;
       return png;
    }
    
    if (setjmp(png.png_ptr->jmpbuf)) {
	MSG_V("PNG Internal error!\n");    
        png_destroy_write_struct(&png.png_ptr, &png.info_ptr);
        fclose(png.fp);
        png.status = ERROR;
        return png;
    }
    
    png.fp = fopen (fname, "wb");
    if (png.fp == NULL) {
	MSG_ERR("\nPNG Error opening %s for writing!\n", strerror(errno));
       	png.status = ERROR;
       	return png;
    }	    
    
    MSG_V("PNG Init IO\n");
    png_init_io(png.png_ptr, png.fp);

    /* set the zlib compression level */
    png_set_compression_level(png.png_ptr, z_compression);

    png_set_IHDR(png.png_ptr, png.info_ptr, image_width, image_height,
       8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
       PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
    
    MSG_V("PNG Write Info\n");
    png_write_info(png.png_ptr, png.info_ptr);
    
    if(cspace) {
    	MSG_V("PNG Set BGR Conversion\n");
    	png_set_bgr(png.png_ptr);
    }	

    png.status = OK;
    return png;
    
}    
       
static uint8_t destroy_png(struct pngdata png) {
	    
    MSG_V("PNG Write End\n");
    png_write_end(png.png_ptr, png.info_ptr);

    MSG_V("PNG Destroy Write Struct\n");
    png_destroy_write_struct(&png.png_ptr, &png.info_ptr);
    
    fclose (png.fp);

    return 0;
}

static void flip_page (unsigned idx)
{
    struct pngdata png;
    png_byte *row_pointers[image_height];
    unsigned k, bppmul = (bpp+7)/8;
    char buf[100];
    UNUSED(idx);
    snprintf (buf, 100, "%08d.png", ++framenum);

    png = create_png(buf);

    if(png.status){
	    MSG_ERR("PNG Error in create_png\n");
    }	     

    MSG_V("PNG Creating Row Pointers\n");
    for ( k = 0; k < image_height; k++ ) row_pointers[k] = &image_data[image_width*k*bppmul];
    
    MSG_V("PNG Writing Image Data\n");
    png_write_image(png.png_ptr, row_pointers);

    destroy_png(png);
}

static uint32_t draw_slice(vo_slice_t *slice)
{
/*
    When encoding input samples that have a sample depth that cannot be
    directly represented in PNG, the encoder must scale the samples up to a
    sample depth that is allowed by PNG.
*/
    uint8_t *dst[3]= {image_data, NULL, NULL};
    int dstStride[3];
    dstStride[0]=image_width*3;
    dstStride[1]=
    dstStride[2]=dstStride[0]>>1;
    sws->swScale(sws,slice->src,slice->stride,slice->y,slice->h,dst,dstStride);
    return 0;
}

static uint32_t query_format(vo_query_fourcc_t *format)
{
    UNUSED(format);
    return 1; /* due swsScaler */
}

static void uninit(void)
{
	if(image_data){ free(image_data);image_data=NULL;}
	if(sws) freeSwsContext(sws);
}

static uint32_t preinit(const char *arg)
{
    if(arg) 
    {
	MSG_ERR("PNG Unknown subdevice: %s\n",arg);
	return ENOSYS;
    }
    return 0;
}

static uint32_t control(uint32_t request, void *data, ...)
{
  switch (request) {
  case VOCTRL_QUERY_FORMAT:
    return query_format((vo_query_fourcc_t*)data);
  case VOCTRL_DRAW_SLICE:
    return draw_slice((vo_slice_t*)data);
  }
  return VO_NOTIMPL;
}
